__all__ = ["VolSeg2dTrainer", "VolSeg2DPredictionManager"]

from volume_segmantics.model.operations.vol_seg_2d_trainer import VolSeg2dTrainer
from volume_segmantics.model.operations.vol_seg_prediction_manager import (
    VolSeg2DPredictionManager,
)
